;*---------------------------------------------------------------------*/
;*    Copyright (c) 1993 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*---------------------------------------------------------------------*/
;*    serrano/prgm/project/bigloo/comptime1.3/Inline/call.scm ...      */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Mar 22 17:07:39 1993                          */
;*    Last change :  Fri Jul  2 13:10:52 1993  (serrano)               */
;*                                                                     */
;*    L'inlining d'un funcall                                          */
;*---------------------------------------------------------------------*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module inline_call
   (include "Var/variable.sch"
	    "Tools/trace.sch")
   (import  engine_param
	    tools_error
	    tools_shape
	    tools_beta
	    tools_alpha
	    tools_progn
	    effect_property
	    scan_lexical
	    scan_tree
	    heap_abstract
	    inline_walk
	    type_misc)
   (export  (inline-function? variable level)
	    (inline-call      form     level)))

;*---------------------------------------------------------------------*/
;*    inline-call ...                                                  */
;*---------------------------------------------------------------------*/
(define (inline-call exp level)
   (trace inline "J'inline l'appel: " (shape exp) #\Newline)
   (let* ((var       (car exp))
	  (function  (if (global? var)
			 (global-value var)
			 (local-value var))))
      ;; il faut avant toutes choses s'assurer que la version de body
      ;; qu'on a pour cette fonction correspond bien a un arbre de
      ;; syntaxe abstraite. Si ce n'est pas le cas, on le construit.
      (if (and (global? var)
	       (eq? (global-import var) 'import)
	       (and (boolean? (function-inline? function))
		    (function-inline? function))
	       (not (eq? (global-info var) '_tree-done_)))
	  ;; voila, on est dans le mauvais cas, on construit l'arbre
	  ;; du code a inline
	  (begin
	     (global-info-set! var '_tree-done_)
	     (make-function-tree var
				 (function-args function)
				 (function-body function))))
      (let* ((args      (function-args function))
	     (body      (normalize-progn (cdr (function-body function))))
	     (subsitute (quoi->par var (function-arity function)
				   args (cdr exp)))
	     (quoi.par  (car subsitute))
	     (bindings  (cdr subsitute)))
	 (trace inline "inline-call: " (shape exp) #\Newline
		"  subsitute: " (shape subsitute) #\Newline)
	 (let* ((body1 (do-inline (beta-reduce (alphatize body) quoi.par)
				  (+fx level 1)))
		(body  (let/O2 body1)))
	    (if (null? bindings)
		body
		`(let ,(map (lambda (f.a)
			       (set-car! (cdr f.a)
					 (do-inline (cadr f.a) (+fx 1 level)))
			       f.a)
			    bindings)
		    ,body))))))
   
;*---------------------------------------------------------------------*/
;*    quoi->par ...                                                    */
;*    -------------------------------------------------------------    */
;*    On calcule la liste des remplacement a effectuer.                */
;*    Cette fonction nous oblige a verifier maintenant l'arite des     */
;*    fonction inlinees. On est oblige de faire ce test de type avant  */
;*    la passe normalement consacre a cela.                            */
;*---------------------------------------------------------------------*/
(define (quoi->par called arity formals actuals)
   (trace inline
	  "quoi->par: " (shape called) " arity: " arity #\Newline
	  "  formals: " (shape formals) #\Newline
	  "  actuals: " (shape actuals) #\Newline)
   (define (fx-quoi->par)
      (let loop ((formals   formals)
		 (actuals   actuals)
		 (quoi.par  '())
		 (bindings  '()))
	 (cond
	    ((null? formals)
	     (if (null? actuals)
		 (cons quoi.par bindings)
		 (begin
		    (partial-error ""
				   "Too many arguments provided in a call to"
				   (shape called))
		    (cons '() '()))))
	    ((null? actuals)
	     (partial-error ""
			    "Too few argumenst provided in a call to"
			    (shape called))
	     (cons '() '()))
	    (else
	     (let ((formal (car formals))
		   (actual (car actuals)))
		(if (can-be-folded? formal actual)
		    ;; on ne cree pas une nouvelle variable.
		    (loop (cdr formals)
			  (cdr actuals)
			  (cons `(,formal . ,actual) quoi.par)
			  bindings)
		    (let ((new (copy-local-variable formal)))
		       (loop (cdr formals)
			     (cdr actuals)
			     (cons `(,formal . ,new) quoi.par)
			     (cons `(,new ,actual)
				   bindings)))))))))
   (define (va-quoi->par)
      (let loop ((formals   formals)
		 (actuals   actuals)
		 (quoi.par  '())
		 (bindings  '())
		 (counter   arity))
	 (cond
	    ((=fx counter -1)
	     (let ((new (copy-local-variable (car formals))))
		(cons (cons `(,(car formals) . ,new) quoi.par)
		      (cons `(,new ,(make-args-list actuals))
			    bindings))))
	    ((null? actuals)
	      (partial-error ""
			    "Too few arguments provided in a call to"
			    (shape called))
	      (cons '() '()))
	    (else
	     (let ((formal (car formals))
		   (actual (car actuals)))
		(if (can-be-folded? formal actual)
		    ;; on ne cree pas une nouvelle variable.
		    (loop (cdr formals)
			  (cdr actuals)
			  (cons `(,formal . ,actual) quoi.par)
			  bindings
			  (+fx counter 1))
		    (let ((new (copy-local-variable formal)))
		       (loop (cdr formals)
			     (cdr actuals)
			     (cons `(,formal . ,new) quoi.par)
			     (cons `(,new ,actual)
				   bindings)
			     (+fx counter 1)))))))))
   (if (<fx arity 0)
       (va-quoi->par)
       (fx-quoi->par)))

;*---------------------------------------------------------------------*/
;*    can-be-folded? ...                                               */
;*    peut-on faire du constante folding en remplacant `formal' par    */
;*    `actual' ?                                                       */
;*---------------------------------------------------------------------*/
(define (can-be-folded? formal actual)
   (or (and (not (pair? actual))
	    (not (global? actual))
	    (or (not (local? actual))
		(not (eq? (local-access actual) 'write)))
	    (not (eq? (local-access formal) 'write)))
       (and (pair? actual)
	    (eq? (car actual) 'quote)
	    (pair? (cdr actual))
	    (symbol? (cadr actual)))))

;*---------------------------------------------------------------------*/
;*    make-args-list ...                                               */
;*---------------------------------------------------------------------*/
(define (make-args-list actuals)
   ;; on construit (en syntaxe abstraite) une serie de cons.
   (let loop ((actuals actuals))
      (if (null? actuals)
	  (abstract-nil)
	  (abstract-cons (car actuals) (loop (cdr actuals))))))

;*---------------------------------------------------------------------*/
;*    inline-function? ...                                             */
;*---------------------------------------------------------------------*/
(define (inline-function? variable level)
   (trace inline "Faut-il inliner " (shape variable) " ? " #\Newline)
   (if (not (global? variable))
       #f
       (let ((function (global-value variable))
	     (import   (global-import variable))
	     (class    (global-class variable)))
	  (if (not (eq? class 'function))
	      #f
	      (let ((body (function-body function)))
		 (trace inline
			"   inlined?: " (function-inline? function) #\Newline
			"       body: " (shape body)
			#\Newline)
		 (cond
		    ((> level *inline-level-max*)
		     (and (pair? body)
			  (eq? (car body) 'begin)
			  (global? (cadr body))
			  (eq? (global-class (cadr body)) 'foreign)))
		    ((boolean? (function-inline? function))
		     (function-inline? function))
		    ((or (<fx *optim* 1)
			 (not *inlining?*))
		     (trace inline "   NON.1" #\Newline)
		     #f)
		    ((eq? import 'import)
		     (trace inline "   NON.2" #\Newline)
		     #f)
		    ((not (pair? body))
		     #t)
		    (else
		     (if (<-nb-leaf-and-no-bind-exit?
			  body
			  (if (<fx (function-arity function) 0)
			      ;; si la fonction est a arite variable,
			      ;; on inline si le corps est plus petit
			      ;; petit que 2 appels a cette fonction
			      (*fx 3 (+fx (+fx 1 (- (function-arity function)))
					  *optim*))
			      ;; dans le cas d'une arite fixe on accepte
			      ;; d'inliner si le corps est plus petit
			      ;; que 3 appels a cette fonction
			      (*fx 3 (+fx (+fx 1 (function-arity function))
					  *optim*))))
			 (begin
			    (trace inline "   OUI" #\Newline)
			    (function-inline?-set! function #t)
			    #t)
			 (begin
			    (trace inline "   NON.3" #\Newline)
			    (function-inline?-set! function #f)
			    #f)))))))))

;*---------------------------------------------------------------------*/
;*    <-nb-leaf-and-no-bind-exit? ..                                   */
;*    -------------------------------------------------------------    */
;*    Est-ce que le nombre de feuille de `body' est plus petit que     */
;*    'max' ou est-ce qu'il y a un bind-exit dans le code ?            */
;*    -------------------------------------------------------------    */
;*    Je regarde s'il y a des bind-exit car il est penible d'integrer  */
;*    une fonction qui en contient car en generale cette forme a       */
;*    tendance a franchement grossir et generer bon nombre de fonction.*/
;*    Je prefere donc etre petit joueur et ne pas faire l'integration  */
;*    dans ce cas.                                                     */
;*---------------------------------------------------------------------*/
(define (<-nb-leaf-and-no-bind-exit? exp max)
   (let loop ((exp exp)
	      (acc 0))
      (cond
	 ((not acc)
	  #f)
	 ((>fx acc max)
	  #f)
	 ((null? exp)
	  acc)
	 ((not (pair? exp))
	  (if (or (eq? exp 'bind-exit) (>=fx acc max))
	      #f
	      (+fx acc 1)))
	 (else
	  (let liip ((exp exp)
		     (acc acc))
	     (cond
		((not acc)
		 #f)
		((>fx acc max)
		 #f)
		((null? exp)
		 acc)
		((not (pair? exp))
		 (loop exp (+fx acc 1)))
		(else
		 (liip (cdr exp) (loop (car exp) acc)))))))))
	      
;*---------------------------------------------------------------------*/
;*    let/O2 ...                                                       */
;*    -------------------------------------------------------------    */
;*    Si on est en mode -O2 et que l'inline n'est pas dangeureuse, on  */
;*    met le resultat dans une variable.                               */
;*---------------------------------------------------------------------*/
(define (let/O2 body)
   (if (and (>=fx *optim* 2)
	    (pair? body)
	    (not (side-effect? body))
	    (bigloo-type? body))
       ;; c'est bon, on utilise une variable.
       (let* ((name (cond
		       ((global? (car body))
			(global-name (car body)))
		       ((local? (car body))
			(local-name (car body)))
		       (else
			'aux)))
	      (aux (cdr (car (allocate-local-variables (list name))))))
	  `(let ((,aux ,body))
	      ,aux))
       body))

       
 
       
 
