;*---------------------------------------------------------------------*/
;*    Copyright (c) 1993 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*---------------------------------------------------------------------*/
;*    .../foreign.scm ...                                              */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Tue Jun 23 18:22:35 1992                          */
;*    Last change :  Thu Mar 18 16:58:06 1993  (serrano)               */
;*                                                                     */
;*    On parse les clauses foreigns                                    */
;*---------------------------------------------------------------------*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module parse_foreign
   (import engine_param
	   type_name
	   var_declare)
   (export (parse-foreign foreigns)))

;*---------------------------------------------------------------------*/
;*    parse-foreign ...                                                */
;*---------------------------------------------------------------------*/
(define (parse-foreign foreigns)
   (let loop ((foreigns foreigns))
      (if (null? foreigns)
	  'done
	  (begin
	     (foreign-parser (car foreigns))
	     (loop (cdr foreigns))))))
   
;*---------------------------------------------------------------------*/
;*    foreign-parser ...                                               */
;*---------------------------------------------------------------------*/
(define (foreign-parser foreign)
      (match-case foreign
      ((include ?string)
       (if (not (string? string))
	   (error "foreign-parser" "Illegal `#include' form" foreign)
	   (set! *include-foreign* (cons string *include-foreign*))))
      ((define ?type ?l-name ?proto ?c-name)
       (if (not (and (or (pair? proto) (null? proto))
		     (string? c-name)
		     (symbol? type)
		     (symbol? l-name)))
	   (error "foreign-parser" "Illegal `define' form" foreign)
	   (declare-global-foreign-function! l-name
					     c-name
					     (user-c-type->type type)
					     (map user-c-type->type proto)
					     'macro-function)))
      ((define ?type ?l-name ?c-name)
       (if (not (and (string? c-name)
		     (symbol? type)
		     (symbol? l-name)))
	   (error "foreign-parser" "Illegal `define' form" foreign)
	   (declare-global-foreign-variable! l-name
					     c-name
					     (user-c-type->type type)
					     'macro-cnst)))
      ((?type ?l-name ?proto ?c-name)
       (if (not (and (or (pair? proto) (null? proto))
		     (string? c-name)
		     (symbol? type)
		     (symbol? l-name)))
	   (error "foreign-parser" "Illegal foreign function form" foreign)
	   (declare-global-foreign-function! l-name
					     c-name
					     (user-c-type->type type)
					     (map user-c-type->type proto)
					     'function)))
      ((?type ?l-name ?c-name)
       (if (not (and (string? c-name)
		     (symbol? l-name)))
	   (error "foreign-parser" "Illegal foreign variable form" foreign)
	   (declare-global-foreign-variable! l-name
					     c-name
					     (user-c-type->type type)
					     'variable)))
      (else
       (error "foreign-parser" "Illegal foreign form" foreign))))
