;*---------------------------------------------------------------------*/
;*    Copyright (c) 1993 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*---------------------------------------------------------------------*/
;*    serrano/prgm/project/bigloo/comptime1.2/Scan/walk.scm ...        */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Mar 18 19:30:04 1993                          */
;*    Last change :  Tue Apr 27 11:56:18 1993  (serrano)               */
;*                                                                     */
;*    Les taches effectuees par cette passe sont:                      */
;*       - trouver toutes les formes statiques                         */
;*       - verifier que chaque prototype a trouve sa definition.       */
;*       - on verifie qu'on ne viole pas les prototype.                */
;*       - Regrouper toutes les formes `top-level'.                    */
;*       - Construire l'arbre de syntaxe abstraite.                    */
;*---------------------------------------------------------------------*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module scan_walk
   (include "Var/variable.sch")
   (import  tools_speek
	    tools_trace
	    tools_error
	    tools_progn
	    tools_shape
	    var_declare
	    var_env
	    engine_param
	    scan_definition
	    scan_temporary
	    scan_tree
	    scan_lexical
	    scan_eval)
   (export  (scan-walk code)))

;*---------------------------------------------------------------------*/
;*    scan-walk ...                                                    */
;*---------------------------------------------------------------------*/
(define (scan-walk code)
   (verbose "   . Scan" #\Newline)
   (start-partial-error "Scan")
   (let ((*top-level* (declare-global-procedure! (initialisation-name)
						 '()
						 *module-name*
						 'export)))
      (let loop  ((code        code)
		  (definitions '()))
	 (if (null? code)
	     ;; on a calculer toutes les formes, il faut maintenant
	     ;; construire l'arbre definitif
	     (make-final-code *top-level* (cons 0 definitions))
	     (match-case (car code)
		((begin . ?body)
		 (loop (append body (cdr code))
		       definitions))
		((or (and (not (?- . ?-)) ?exp)
		     (begin . ?body))
		 (loop (cdr code)
		       (cons (make-expression-tree (car code) 'read '())
			     definitions)))
		((or (define-inline (?name . ?-) ?-)
		     (define (?name . ?-) ?-))
		 (loop (cdr code)
		       (cons (scan-function-definition (car code))
			     definitions)))
		((define ?var ?val)
		 (loop (cdr code)
		       (cons (scan-value-definition (car code))
			     definitions)))
		(else
		 (loop (cdr code)
		       (cons (make-expression-tree (car code) 'read '())
			     definitions))))))))
   
;*---------------------------------------------------------------------*/
;*    make-final-tree ...                                              */
;*    -------------------------------------------------------------    */
;*    Le travail essentiel de cette fonction est de repartir en deux   */
;*    categories les definitions. Il y a celles qui sont effectivement */
;*    des definitions et il y a celles qui doivent etre mises dans la  */
;*    fonction d'initialisation.                                       */
;*---------------------------------------------------------------------*/
(define (make-final-code top-level definitions)
   ;; on a lu toutes les variables temporaires, il faut
   ;; maintenant les mettre dans *Genv*
   (fixe-temporary!)
   ;; maintenant qu'on a tout vu, on peut s'arretter s'il
   ;; y a des erreurs.
   (fail-if-partial-error)
   ;; on verifie que chaque global a trouve sa definition
   ;; on supprime la liste la fonction d'initialisation
   (remove-from-remember-list! (initialisation-name))
   (if (not (null? (get-remember-list)))
       (error "scan"
	      "some declared variable aren't defined"
	      (get-remember-list)))
   (let loop ((definitions definitions)
	      (def-forms  '())
	      (init-forms '()))
      (if (null? definitions)
	  ;; on construit la fonction d'initialisation et on retourne
	  (let ((res (cons (make-init-function top-level init-forms)
			   def-forms)))
	     (when-trace '(env init)
			 (lambda ()
			    (fprint *trace-port* "==> Genv (scan): ")
			    (pp-global-environment *trace-port*
						   *Genv*)))
	     (when-trace 'init
			 (lambda ()
			    (fprint *trace-port* #\Newline
				    "_________________________________"
				    #\Newline "==> Tree: ")
			    (for-each (lambda (def)
					 (fprint *trace-port*
						 (shape def)
						 #\Newline))
				      res)))
	     res)
	  (match-case (car definitions)
	     ((define ?global (lambda ?args ?body))
	      (if (eq? (global-class global) 'function)
		  (let ((value (global-value global)))
		     (function-body-set! value body)
		     (function-args-set! value args)
		     (loop (cdr definitions)
			   (cons global def-forms)
			   init-forms))
		  (loop (cdr definitions)
			def-forms
			(cons (define->set! top-level (car definitions))
			      init-forms))))
	     (else
	      (loop (cdr definitions)
		    def-forms
		    (cons (car definitions)
			  init-forms)))))))

;*---------------------------------------------------------------------*/
;*    define->set! ...                                                 */
;*---------------------------------------------------------------------*/
(define (define->set! top-level form)
   (let* ((var   (cadr form))
	  (args  (cadr (caddr form)))
	  (body  (caddr (caddr form)))
	  (name  (get-new-lambda-name (function-arity (global-value var))))
	  (local (cdar (allocate-local-functions (list name)))))
      ;; on prend l'arite de la fonction et ...
      (function-arity-set! (local-value local)
			   (function-arity (global-value var)))
      ;; on efface le champs valueur qui ne sert plus a rien.
      (global-value-set! var '())
      `(set! ,var
	     (labels ((,local ,args ,body))
		,local))))
		 
;*---------------------------------------------------------------------*/
;*       make-init-function ...                                        */
;*---------------------------------------------------------------------*/
(define (make-init-function top-level forms)
   (let ((forms (append (eval-initialisation) forms)))
      (function-body-set! (global-value top-level) (normalize-progn forms))
      (function-args-set! (global-value top-level) '())
      top-level))

