;*---------------------------------------------------------------------*/
;*    Copyright (c) 1993 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*---------------------------------------------------------------------*/
;*    .../application.scm ...                                          */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Mar 24 17:11:36 1993                          */
;*    Last change :  Fri Jun 25 15:04:34 1993  (serrano)               */
;*                                                                     */
;*    On type une application (c'est la que ca se corse...)            */
;*---------------------------------------------------------------------*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module type_application
   (include "Var/variable.sch"
	    "Tools/trace.sch")
   (import  type_enforce
	    type_cast
	    type_expression
	    type_type
	    tools_error
	    tools_shape
	    heap_abstract)
   (export  (type-atom-application exp to)
	    (type-unknown-bigloo-application exp to)))

;*---------------------------------------------------------------------*/
;*    type-atom-application ...                                        */
;*---------------------------------------------------------------------*/
(define (type-atom-application exp to)
   (let ((fun (car exp))
	 class
	 value)
      (cond
	 ((global? fun)
	  (set! class (global-class fun))
	  (set! value (global-value fun)))
	 ((local? fun)
	  (set! class (local-class fun))
	  (set! value (local-value fun)))
	 (else
	  (partial-error "" "Illegal application" (shape exp))))
      (cond
	 ((eq? class 'foreign)
	  (type-foreign-application exp to))
	 ((eq? class 'function)
	  (type-known-bigloo-application exp value to))
	 ((eq? class 'return)
	  (type-return-application exp value to))
	 (else
	  (type-unknown-bigloo-application exp to)))))

;*---------------------------------------------------------------------*/
;*    type-known-bigloo-application ...                                */
;*    -------------------------------------------------------------    */
;*    Ici, on se contente de verifier que les arites sont correctes    */
;*    et eventuellement construire la liste d'appelle pour les         */
;*    fonctions a arite multiple.                                      */
;*---------------------------------------------------------------------*/
(define (type-known-bigloo-application exp fun to)
   (if (>=fx (function-arity fun) 0)
       (type-known-bigloo-fx-application exp fun to)
       (type-known-bigloo-va-application exp fun to)))

;*---------------------------------------------------------------------*/
;*    type-known-fx-application ...                                    */
;*---------------------------------------------------------------------*/
(define (type-known-bigloo-fx-application exp fun to)
   (if (not (=fx (function-arity fun) (length (cdr exp))))
       (partial-error "" "Wrong number of argument for" (shape (car exp)))
       (let loop ((hook (cdr exp)))
	  (if (null? hook)
		 ;; on n'a pas besoin de tester le `car' de `exp' car on
	      ;; sait deja que c'est une variable qui est liee a une
	      ;; fonction.
	      (cast exp 'bobj to)
	      (begin
		 (set-car! hook (type-expression (car hook) 'bobj))
		 (loop (cdr hook)))))))

;*---------------------------------------------------------------------*/
;*    type-known-bigloo-va-application ...                             */
;*---------------------------------------------------------------------*/
(define (type-known-bigloo-va-application exp fun to)
   ;; attention, c'est express que je mesure exp et non pas (cdr exp),
   ;; c'est pour m'eviter de rajouter 1 !! :-D
   (if (>fx (negfx (length exp)) (function-arity fun))
       (partial-error "" "Too few arguments provided" (shape (car exp)))
       (let loop ((hook    exp)
		  (counter (function-arity fun)))
	  (if (=fx counter -1) 
	      (let ((new (make-args-list (cdr hook))))
		 (if (null? (cdr hook))
		     (set-cdr! hook (cons new '()))
		     (begin
			(set-car! (cdr hook) new)
			(set-cdr! (cdr hook) '())))
		 (cast exp 'bobj to))
	      (begin
		 (set-car! (cdr hook) (type-expression (cadr hook) 'bobj))
		 (loop (cdr hook)
		       (+fx counter 1)))))))
	 
;*---------------------------------------------------------------------*/
;*    make-args-list ...                                               */
;*---------------------------------------------------------------------*/
(define (make-args-list actuals)
   ;; on construit (en syntaxe abstraite) une serie de cons.
   (let loop ((actuals actuals))
      (if (null? actuals)
	  (abstract-nil)
	  (abstract-cons (type-expression (car actuals) 'bobj)
			 (loop (cdr actuals))))))

;*---------------------------------------------------------------------*/
;*    type-return-application ...                                      */
;*---------------------------------------------------------------------*/
(define (type-return-application exp fun to)
   (if (not (=fx 1 (length (cdr exp))))
       (partial-error "" "Wrong number of argument for" (shape (car exp)))
       (let loop ((hook (cdr exp)))
	  (if (null? hook)
	      (begin
		 (set-cdr! exp (cons (car exp) (cdr exp)))
		 (set-car! exp 'return-from)
		 (cast exp 'bobj to))
	      (begin
		 (set-car! hook (type-expression (car hook) 'bobj))
		 (loop (cdr hook)))))))
   
;*---------------------------------------------------------------------*/
;*    type-unknown-bigloo-application ...                              */
;*    -------------------------------------------------------------    */
;*    On pose les tests verfiant que c'est bien une fonction et on     */
;*    fait un appel calcule                                            */
;*    -------------------------------------------------------------    */
;*    On passe toujours un argument supplementaire lors d'un appel     */
;*    calcule pour que les va_list puisse etre examinee en faisant     */
;*    un `*++'.                                                        */
;*---------------------------------------------------------------------*/
(define (type-unknown-bigloo-application exp to)
   (trace type "type-unknown-bigloo-application: " (shape exp) #\Newline)
   (if (null? (cdr exp))
       (begin
	  (set-car! exp (enforce-procedure (abstract-cint->bint
					    (length (cdr exp)))
					   (type-expression (car exp)
							    'bobj)))
	  (set-cdr! exp (cons (abstract-eoa) '()))
	  (cast (cons 'funcall exp) 'bobj to))
       (let loop ((hook (cdr exp)))
	  (if (null? (cdr hook))
	      (begin
		 (set-car! hook (type-expression (car hook) 'bobj))
		 (set-car! exp (enforce-procedure (abstract-cint->bint
						   (length (cdr exp)))
						  (type-expression (car exp)
								   'bobj)))
		 (set-cdr! hook (cons (abstract-eoa) '()))
		 (cast (cons 'funcall exp) 'bobj to))
	      (begin
		 (set-car! hook (type-expression (car hook) 'bobj))
		 (loop (cdr hook)))))))

;*---------------------------------------------------------------------*/
;*    type-foreign-application ...                                     */
;*    -------------------------------------------------------------    */
;*    On pose les tests de verfication de type et on pose les casts.   */
;*---------------------------------------------------------------------*/
(define (type-foreign-application exp to)
   (let* ((fun       (car exp))
	  (foreign   (global-value fun)))
      (if (not (or (eq? (foreign-class foreign) 'function)
		   (eq? (foreign-class foreign) 'macro-function)))
	  (begin
	     (partial-error ""
			    "Illegal application of foreign variable"
			    (shape fun))
	     exp)
	  (let ((type-res  (cadr (foreign-type foreign)))
		(type-args (cddr (foreign-type foreign))))
	     (let loop ((hook  (cdr exp))
			(types type-args))
		(cond
		   ((null? hook)
		    (if (null? types)
			(cast exp type-res to)
			(partial-error
			 ""
			 "Too few arguments provided to foreign function"
			 (shape fun))))
		   ((null? types)
		    (partial-error
		     ""
		     "Too many arguments provided to foreign function"
		     (shape fun))
		    exp)
		   (else
		    (begin
		       (set-car! hook (type-expression (car hook) (car types)))
		       (loop (cdr hook)
			     (cdr types))))))))))


