/*---------------------------------------------------------------------*/
/*    Copyright (c) 1993 by Manuel Serrano. All rights reserved.       */
/*                                                                     */
/*                                     ,--^,                           */
/*                               _ ___/ /|/                            */
/*                           ,;'( )__, ) '                             */
/*                          ;;  //   L__.                              */
/*                          '   \    /  '                              */
/*                               ^   ^                                 */
/*                                                                     */
/*                                                                     */
/*    This program is distributed in the hope that it will be useful.  */
/*    Use and copying of this software and preparation of derivative   */
/*    works based upon this software are permitted, so long as the     */
/*    following conditions are met:                                    */
/*           o credit to the authors is acknowledged following         */
/*             current academic behaviour                              */
/*           o no fees or compensation are charged for use, copies,    */
/*             or access to this software                              */
/*           o this copyright notice is included intact.               */
/*      This software is made available AS IS, and no warranty is made */
/*      about the software or its performance.                         */
/*                                                                     */
/*      Bug descriptions, use reports, comments or suggestions are     */
/*      welcome Send them to                                           */
/*        <Manuel.Serrano@inria.fr>                                    */
/*        Manuel Serrano                                               */
/*        INRIA -- Rocquencourt                                        */
/*        Domaine de Voluceau, BP 105                                  */
/*        78153 Le Chesnay Cedex                                       */
/*        France                                                       */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*    serrano/prgm/project/bigloo/runtime1.3/Clib/main.c ...           */
/*                                                                     */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Jul 17 09:40:49 1992                          */
/*    Last change :  Tue Jun 22 15:23:58 1993  (serrano)               */
/*                                                                     */
/*    Le fichier de main de toute application. Comme je m'y prends     */
/*    plus intelligement que dans la version 0.8 (si, si :-), je       */
/*    peux compiler une fois pour toute ce fichier. Je ne suis plus    */
/*    oblige de le construire et le compiler pour chaque application.  */
/*---------------------------------------------------------------------*/
#include <bigloo.h>
#if( defined( sun ) || defined( sony_news ) )
#   include <machine/vmparam.h>
#endif

/*---------------------------------------------------------------------*/
/*    Quelques macros de calcul de taille memoire pour les GCs         */
/*---------------------------------------------------------------------*/
#define MegToByte(x) ((x) * (1024 * 1024))
#define MegToKilo(x) ((x) * 1024)
#define MegTo4K(x)   ((x) * (1024 >> 2))

#define DEFAULT_HEAP_SIZE 8

/*---------------------------------------------------------------------*/
/*    Des recuperations externes                                       */
/*---------------------------------------------------------------------*/
extern void dprint();
extern void cref();
extern void memshow();
extern char *getenv();

/*---------------------------------------------------------------------*/
/*    Une variable pour memoriser le bas de la pile                    */
/*---------------------------------------------------------------------*/
char *stack_bottom;
int   glob_dummy;

/*---------------------------------------------------------------------*/
/*    bigloo_exit ...                                                  */
/*---------------------------------------------------------------------*/
obj_t
bigloo_exit( val )
obj_t val;
{
   FREE_ALLOCATION();
   exit( (int)INTEGERP( val ) ? (int)CINT( val ) : 0 );
}

/*---------------------------------------------------------------------*/
/*    bigloo_main ...                                                  */
/*    -------------------------------------------------------------    */
/*    Le vrai point d'entree d'une application `bigloo'. La fonction   */
/*    `main' qui est cree par le compilo ne sert a rien. Elle existe   */
/*    uniquement car je ne suis pas que tous les compilos supporte que */
/*    la fonction C `main' soit mise en librairie. Comme ca, je suis   */
/*    tranquile.                                                       */
/*---------------------------------------------------------------------*/
_bigloo_main( argc, argv )
int   argc;
char *argv[];
{
   int   heap_size;
   int   mega_size;
   char *env_size;
   obj_t cons;
   int   i;

   /* dummy */
   if( argc < 1 )
      dprint( argc );

   /* on memorise l'adresse du bas de la pile */
   stack_bottom = ( (char *)&argc - 4);
   
   /* on initialise le tas */
   if( !(env_size = getenv( "BIGLOOHEAP" )) )
      mega_size = DEFAULT_HEAP_SIZE;
   else
      mega_size = atoi( env_size );

#if( (GC == BOEHM_1_X_GC) || (GC == BOEHM_2_X_GC) )
   heap_size = MegTo4K( mega_size );
#else
   heap_size = MegToByte( mega_size );
#endif

   if( !INIT_ALLOCATION( heap_size ) )
   {
      char mes[ 600 ];

      sprintf( mes, "%s Meg wanted", heap_size / (1024 * 1024) );
      c_error( "Can't allocate heap", mes, EHEAP );
   }
   else
   {
      /* on initialise les objects pre-alloues (les chars, bool, ...) */
      init_objects();

      /* on construit la liste des argv */
      for( i = argc - 1, cons = BNIL; i >= 0; i-- )
         cons = MAKE_PAIR( c_constant_string_to_string( argv[ i ] ), cons );

      /* on appelle le main de l'utilisateur */
      bigloo_exit( bigloo_main( cons ) );
   }
}


